#!/usr/bin/env ruby
lib_path = __FILE__ == '(irb)' ? File.join(Dir.cwd, 'lib') : File.expand_path('../../../lib', __FILE__)
$:.unshift(lib_path) unless $:.include?(lib_path) or !File.exists?(lib_path)

require 'rubygems'
require 'optparse'
require 'open3'
require 'pp'

require 'envoi/mam/agent/cli'
require 'envoi/mam/cantemo/agent'
require 'envoi/mam/cantemo/agent/watch_folder_manager'

Envoi::Mam::Agent::CLI::CONFIG_FILE_PATHS.clear
Envoi::Mam::Agent::CLI::CONFIG_FILE_PATHS.concat [
   './cantemo-portal-agent-config.json',
   '~/cantemo-portal-agent-config.json',
   './.cantemo-portal-agent-config.json',
   '~/.cantemo-portal-agent-config.json',
]
default_config_file_paths = Envoi::Mam::Agent::CLI::CONFIG_FILE_PATHS
current_command = ARGV.shift
# ARGV << '--help' if ARGV.empty?

@args = {
    :config_file_path  => default_config_file_paths,
    :dry_run           => false,
    :operation         => :upload,
    :preserve_path     => false,
    :transfer_type     => '',
}
def args; @args end

op = OptionParser.new
op.on('-c', '--config-file-path FILEPATH', 'The path to the configuration file.',
      "Default Path(s): #{args[:config_file_path]}") { |v| args[:config_file_path] = v }
# op.on('-b', '--[no-]background', 'Tells the application to run in the background.') { |v| args[:daemonize] = v }
op.on_tail('-t', '--log-to FILEPATH', 'Log file to log to') { |v| args[:log_to] = v }
op.on_tail('-l', '--log-level LEVEL', %w(fatal error warn info debug), 'Select logging level') do |v|
  args[:log_level] = v
end
op.on_tail('-v', '--version', 'Output the version and exit.') { puts "Version #{Cantemo::Portal::Agent::VERSION}"; exit }
op.on_tail('-h', '--help', 'Show this message.') { puts op; exit }
op.load
op.parse!

sub_command = (ARGV.first || 'run').downcase
sub_command.tr('-_', '')

if sub_command == 'help'
  puts op
  exit
end

config_file_path = args[:config_file_path]
if config_file_path.is_a?(Array)
  args[:config_file_path].map! { |v| File.expand_path(v) }
  config_file_path = config_file_path.find { |file_path| File.exist?(file_path) }
else
  args[:config_file_path] = config_file_path = File.expand_path(config_file_path)
end

unless config_file_path && File.exist?(config_file_path)
  abort("Configuration file not found. Path(s): #{args[:config_file_path]}")
end

# daemon_control_command_present = sub_command && begin
#   sub_command = sub_command.dup.downcase if sub_command
#   %w(start stop restart zap killall status).include?(sub_command)
# end
# should_daemonize = (daemon_control_command_present && %w(start restart).include?(sub_command)) || args[:daemonize]

WATCH_FOLDER_MANAGER_CLASS = Envoi::Mam::Cantemo::Agent::WatchFolderManager

if sub_command == 'run'
  begin
    WATCH_FOLDER_MANAGER_CLASS.run(args)
  rescue => e
    abort("An exception occurred: #{e.message}#{args[:log_level] == 'debug' ? " - #{e.backtrace.first}" : ''}")
  end
  exit
end

module Application

  module OS
    def OS.windows?;  (/cygwin|mswin|mingw|bccwin|wince|emx/ =~ RUBY_PLATFORM) != nil end
    def OS.mac?;      (/darwin/ =~ RUBY_PLATFORM) != nil end
    def OS.unix?;     !OS.windows? end
    def OS.linux?;    OS.unix? and not OS.mac? end
  end

  if OS.windows?
    require 'win32/daemon'
    class Watcher < Win32::Daemon

      LOG_FILE_PATH = 'C:\\tmp\\win32_daemon_test.log'
      log_file_dir_path = File.dirname(LOG_FILE_PATH)
      Dir.mkdir_p(log_file_dir_path) unless Dir.exist?(log_file_dir_path)

      def write_log(message = nil, &block)
        if block_given?
          begin
            f = File.open(LOG_FILE_PATH, 'a')
            yield f
          ensure
            f.close if f && f.open?
          end
        else
          File.open(LOG_FILE_PATH, 'a') { |f| f.puts(message) }
        end
      end

      def service_init
        @process_thread = nil
      end

      def service_main(*args)
        msg = 'service_main entered at: ' + Time.now.to_s

        write_log{ |f|
          f.puts msg
          f.puts "Args: " + args.join(',')
        }

        @watch_folder_manager = nil
        @process_thread = Thread.new(@watch_folder_manager, args) do |watch_folder_manager, args|
          watch_folder_manager = WATCH_FOLDER_MANAGER_CLASS.new(args)
          watch_folder_manager.run
        end

        while running?
          if state == RUNNING
            sleep 20
            msg = 'Service is running as of: ' + Time.now.to_s
            write_log { |f| f.puts msg }
          else # PAUSED or IDLE
            sleep 0.5
          end
        end


        # We've left the loop, the daemon is about to exit.

        write_log{ |f| f.puts "STATE: #{state}" }

        msg = 'service_main left at: ' + Time.now.to_s

        write_log{ |f| f.puts msg }

      end

      def service_stop
        @watch_folder_manager.stop
      end

      def service_pause
        @watch_folder_manager.pause
      end

      def service_resume
        @watch_folder_manager.resume
      end

      # Watcher - Windows
    end
  else
    class Watcher
      APP_NAME = 'cantemo-portal-watch-folders'

      def self.run(args)
        WATCH_FOLDER_MANAGER_CLASS.run(args)
      end

      def self.daemons_run_proc_with_cleanup(options, &block)
        options[:dir_mode] = :normal
        options[:dir]      = File.split(__FILE__)[0]

        if block_given?
          options[:mode] = :proc
          options[:proc] = block
        end

        controller                            = Daemons::Controller.new(options, ARGV)
        _command, _controller_part, _app_part = controller.class.split_argv(ARGV)

        controller_group                 = Daemons::ApplicationGroup.new(controller.app_name, controller.options)
        controller_group.controller_argv = _controller_part
        controller_group.app_argv        = _app_part

        controller_group.setup
        applications = controller_group.applications

        is_running = applications.find { |a| a.running? }
        if !applications.empty?
          puts "Found #{applications.length} existing pid file(s) #{applications.map { |a| a.pid.pid }}"
          should_zap_all = !is_running || (applications.length == 1 && applications.first.pid.pid == 0)
          if should_zap_all
            warn "Found stale pid file(s)"
            controller_group.zap_all
            controller_group.options[:force] = true
            # controller_group.applications = []

            controller.options[:force] = true
          end
        end

        Daemons.run_proc(options[:app_name], options, &block)
        # controller.catch_exceptions do
        #   controller.run
        # end

      end

      def self.run_with_process_manager(args, options = {})
        # ARGV.unshift 'run' unless %w(start stop restart run zap killall status).include? ARGV.first
        require 'daemons'
        proc     = Proc.new { self.run(args) }
        app_name = APP_NAME

        # options[:app_name] = app_name
        # daemons_run_proc_with_cleanup(options, &proc)

        # options[:force] = true
        Daemons.run_proc(app_name, options, &proc)
      end

      def self.run_in_foreground(args, options = { })
        self.run(args)
      end
      # Watcher - *nix
    end
  end


  def run_in_background(args)

  end

  def self.run_in_foreground(args)
    Watcher.run(args)
  end

  def self.run_with_process_manager(args)
    Watcher.run_with_process_manager(args)
  end

  def self.run(args)

  end

end

# puts "ARGV: #{ARGV}"
# exit

if !Application::OS.windows?
  begin
    ARGV.unshift sub_command unless ARGV.first == sub_command
    Application.run_with_process_manager(args)
    exit
  rescue => e
    abort("An exception occurred: #{e.message}#{args[:log_level] == 'debug' ? " - #{e.backtrace.first}" : ''}")
  end
end


module Application

  class WindowsDaemon

  end

  class WindowsService

  end

end


# Options:
#    install    - Installs the service.
#    start      - Starts the service.
#    stop       - Stops the service.
#    pause      - Pauses the service.
#    resume     - Resumes the service.
#    uninstall  - Uninstalls the service.
#    delete     - Same as uninstall.
require 'win32/service'
require 'rbconfig'
include Win32
include RbConfig

SERVICE_NAME = 'cantemo_portal_watch_folder_agent'
SERVICE_DISPLAYNAME = 'Cantemo Portal Watch Folder Agent'

# You must provide at least one argument.
abort('No argument provided.') unless sub_command
abort('Service is not installed.') unless Service.exists?(SERVICE_NAME) || sub_command == 'install'


case sub_command
when 'runasservice'
  Application::Watcher.mainloop
when 'start'
  if Service.status(SERVICE_NAME).current_state != 'running'
    # Service.start(SERVICE_NAME, nil, args)
    Service.start(SERVICE_NAME)
    while Service.status(SERVICE_NAME).current_state != 'running'
      puts 'One moment...' + Service.status(SERVICE_NAME).current_state
      sleep 1
    end
    puts 'Service ' + SERVICE_NAME + ' started'
  else
    puts 'Already running'
  end
when 'stop'
  if Service.status(SERVICE_NAME).current_state != 'stopped'
    Service.stop(SERVICE_NAME)
    while Service.status(SERVICE_NAME).current_state != 'stopped'
      puts 'One moment...' + Service.status(SERVICE_NAME).current_state
      sleep 1
    end
    puts 'Service ' + SERVICE_NAME + ' stopped'
  else
    puts 'Already stopped'
  end
when 'pause'
  if Service.status(SERVICE_NAME).current_state != 'paused'
    Service.pause(SERVICE_NAME)
    while Service.status(SERVICE_NAME).current_state != 'paused'
      puts 'One moment...' + Service.status(SERVICE_NAME).current_state
      sleep 1
    end
    puts 'Service ' + SERVICE_NAME + ' paused'
  else
    puts 'Already paused'
  end
when 'resume'
  if Service.status(SERVICE_NAME).current_state != 'running'
    Service.resume(SERVICE_NAME)
    while Service.status(SERVICE_NAME).current_state != 'running'
      puts 'One moment...' + Service.status(SERVICE_NAME).current_state
      sleep 1
    end
    puts 'Service ' + SERVICE_NAME + ' resumed'
  else
    puts 'Already running'
  end
when 'status'
  puts Service.status(SERVICE_NAME)
when 'install'
  # Quote the full path to deal with possible spaces in the path name.
  ruby = File.join(CONFIG['bindir'], CONFIG['ruby_install_name']).tr('/', '\\')
  path = File.join(CONFIG['bindir'], 'cantemo-portal-agent')
  # cmd = %Q("#{ruby}" "#{path}" runasservice --config-file-path "#{config_file_path}")
  cmd = %Q("#{ruby}" "#{path}" run-as-service)

  Service.new(
      :service_name     => SERVICE_NAME,
      :display_name     => SERVICE_DISPLAYNAME,
      :description      => 'Watch Folder Agent for Cantemo Portal',
      :binary_path_name => cmd
  )
  puts 'Service ' + SERVICE_NAME + ' installed'
when 'uninstall', 'delete'
  if Service.status(SERVICE_NAME).current_state != 'stopped'
    Service.stop(SERVICE_NAME)
  end
  while Service.status(SERVICE_NAME).current_state != 'stopped'
    puts 'One moment...' + Service.status(SERVICE_NAME).current_state
    sleep 1
  end
  Service.delete(SERVICE_NAME)
  puts 'Service ' + SERVICE_NAME + ' deleted'

else
  raise ArgumentError, 'unknown option: ' + sub_command
end


