require 'listen' # https://github.com/guard/listen
require 'pp'
Pathname
module Envoi
  module WatchFolderUtility
    class WatchFolder
      class Handler
        class Listen

          attr_accessor :logger,
                        :definition,
                        :paths,
                        :known_path_map,
                        :ignored_files_map,
                        :last_poll_time,
                        :min_stable_poll_count,
                        :min_stable_time,
                        :poll_interval,

                        :listener, :lock,

                        :quarantine_directory_path

          class DiscoveredPath < OpenStruct

            def min_stable_time
              handler.min_stable_time
            end

            def min_stable_poll_count
              handler.min_stable_poll_count
            end

            def stable?
              return false if deleted?
              return false unless exist?

              _min_stable_time = min_stable_time
              if _min_stable_time
                 ((Time.now - self[:event_timestamp]) >= _min_stable_time)
              else
                (self[:stable_poll_count] >= min_stable_poll_count)
              end
            end

            def exist?
              File.exist?(self[:path])
            end

            def deleted?
              self[:event_type] == :deleted
            end

            def summary
              _table = @table.dup
              _table.delete(:handler)
              _table
            end

            def watch_folder
              self[:watch_folder]
            end

            def watch_folder=(wf)
              self[:watch_folder] = wf
            end

            def handler
              self[:handler]
            end

            def handler=(_handler)
              self[:handler] = _handler
            end

            def ignore
              handler.add_to_ignore(self)
            end

            def ignore?
              handler.ignored_files_map[self[:path]] == self
            end

          end

          def initialize(args = { })
            initialize_logger(args)
            logger.debug { "Initializing Watch Folder Handler '#{self.class.name}'. Args: #{args}" }

            @definition = args[:definition]
            raise ArgumentError, "Definition is a required argument." unless definition

            @known_path_map = { }
            @ignored_files_map = { }

            @lock = Mutex.new

            @processing = { }

            process_definition(@definition)
            initialize_listener
            logger.debug ( "Watch Folder Handler '#{self.class.name}' Initialized. ")
          end

          def initialize_logger(args = { })
            @logger = args[:logger] ||= Logger.new(args[:log_to] || STDOUT)
            log_level = args[:log_level] || Logger::DEBUG
            if log_level
              @logger.level = log_level
              args[:logger] = @logger
            end
            @logger
          end

          def process_definition(watch_folder_def)
            # @paths = [ File.expand_path('~/watch_folders/aspera') ]
            @paths = watch_folder_def['paths'] || []
            path = watch_folder_def['path']
            @paths << path if path && !@paths.include?(path)

            # @quarantine_directory_path = definition['quarantine_directory_path'] || definition['quarantine_path']
            @poll_interval = watch_folder_def['poll_interval'] || 15
            @min_stable_poll_count = watch_folder_def['minimum_stable_poll_count'] || 3
            @min_stable_time = watch_folder_def['minimum_stable_time']
            if @min_stable_time.nil? && @min_stable_poll_count && @min_stable_time
              @min_stable_time = @poll_interval * @min_stable_poll_count
            end
            @processing_limit = watch_folder_def['processing_limit'] || 10
          end

          def process_deleted_path(path)
            return path.each { |path| process_deleted_path(path) } if path.is_a?(Array)
            logger.debug { "PATH #{:deleted} '#{path}'" }
            file = known_path_map[path]
            known_path_map.delete(path)
            ignored_files_map.delete(path)
            return unless file
            file.event_type = :deleted
            file.event_timestamp = Time.now
            file.stable_poll_count = 0
            file.deleted = true
          end

          def process_add_or_modified_path(path, event_type)
            return path.each { |path| process_add_or_modified_path(path, event_type) } if path.is_a?(Array)
            logger.debug { "PATH #{event_type} '#{path}'" }
            # lock.synchronize { known_path_map[path] = OpenStruct.new({ type: event_type, path: path, stable_poll_count: 0, timestamp: Time.now}) }
            file = known_path_map[path] ||= DiscoveredPath.new({ path: path, processed: false, handler: self })
            file.event_type = event_type
            file.event_timestamp = Time.now
            file.stable_poll_count = 0
            file
          end

          def initialize_listener(args = { })
            @listener = ::Listen.to(*paths, {}) do |m, a, r|
              if !r.empty?
                logger.info { "#{r.length} file(s) removed." }
                lock.synchronize { r.each { |f| process_deleted_path(f) } }
              end

              if !m.empty?
                logger.info { "#{m.length} file(s) modified." }
                lock.synchronize { process_add_or_modified_path(m, :modified) }
              end

              if !a.empty?
                logger.info { "#{a.length} file(s) added." }
                lock.synchronize { process_add_or_modified_path(a, :added) }
              end

            end
          end

          # Listen doesn't perform an initial inventory by default so this method will pick up
          # any files that are already present
          def perform_initial_inventory
            _config = @listener.instance_variable_get(:@backend).instance_variable_get(:@adapter).config
            # change_config = ::Listen::Change::Config.new(_config.queue, _config.silencer)

            _snapshots ||= {}
            _config.directories.each do |dir|
              record = ::Listen::Record.new(dir)
              record.build
              record.instance_variable_get(:@tree).each do |rel_path, fstats|
                _path = File.join(dir, rel_path)
                process_add_or_modified_path(_path, :added)
              end
              # snapshot = ::Listen::Change.new(change_config, record)
              # _config.optimize_changes(snapshot)
              # _snapshots[dir] = snapshot
            end

            # pp _snapshots
          end

          def add_to_ignore(file)
            ignored_files_map[file.path] = file
          end

          # Really this is just for debugging as polling isn't used to determine file stability
          def poll(options = { })
            # return if @min_stable_time
            should_increment_counter = options.fetch(:should_increment_counter, true)

            @previous_poll_time = @last_poll_time
            @last_poll_time = Time.now

            lock.synchronize do
              @known_path_map.each do |fp, f|
                next if f.ignore?
                unless f.exist?
                  logger.debug { "Skipping Missing File. #{f.summary}" }
                  next
                end

                logger.debug { "Incrementing Stable Stats: #{f.summary}" }
                f.last_poll_time = Time.now
                f[:stable_poll_count] += 1 if should_increment_counter
              end
            end
          end

          def stable_files
            @known_path_map.select { |p,f| f.stable? }.values
          end

          def run_once(options = { })
            perform_initial_inventory
          end

          def run(options = { })
            logger.info { "Starting handler for #{paths}" }
            perform_initial_inventory
            listener.start
          end

          def stop
            listener.stop if listener && listener.respond_to?(:stop)
          end

          InspectKey = :__inspect_key__ # :nodoc:

          def inspect
            str = "#<#{self.class}"

            ids = (Thread.current[InspectKey] ||= [])
            if ids.include?(object_id)
              return str << ' ...>'
            end

            ids << object_id
            # begin
            #   first = true
            #   for k,v in @table
            #     str << "," unless first
            #     first = false
            #     str << " #{k}=#{v.inspect}"
            #   end
            #   return str << '>'
            # ensure
            #   ids.pop
            # end
          end
          alias :to_s :inspect

        end
      end
    end
  end
end


# watcher = Envoi::WatchFolderUtility::WatchFolder::Handler::Listen.new
# watcher.run
#
#
#
#
# class WatchFolder
#
#   attr_accessor :definition, :state, :client, :watcher,
#                 :previous_poll_time, :last_poll_time
#
#   def initialize(args = { })
#
#   end
#
#   def poll(&block)
#
#   end
#
#   def run_once(&block)
#     poll(&block)
#   end
#
#   def run(poll_interval = 15, &block)
#     loop do
#       begin
#         run_once(&block)
#         sleep poll_interval
#       rescue SystemExit, Interrupt
#         break
#       end
#     end
#   end
#
# end